"""Collection of dynamic system creation classes for control problems."""

import numpy
import toolz
import networkx


# Simple Systems

@toolz.curry
def harmonic_oscillator(actuator, omega=1.0):
    """Return harmonic oscillator with actuator built in.

    :param actuator: callable(*y, *args).
    :param omega: angular frequency of the oscillator.
    """
    def dy(t, y, *args):
        dy0 = y[1]
        dy1 = -omega**2 * y[0] + actuator(*y, *args)
        return [dy0, dy1]
    return dy


@toolz.curry
def anharmonic_oscillator(actuator, omega=1.0, c=1.0, k=1.0):
    """Return anharmonic oscillator with actuator built in."""
    def dy(t, y, *args):
        dy0 = y[1]
        dy1 = - omega**2 * y[0] - k * y[0]**2 - c * y[1] + actuator(*y, *args)
        return [dy0, dy1]
    return dy


@toolz.curry
def lorenz_in_3(actuator, s=10.0, r=28.0, b=8.0 / 3.0):
    """Return lorenz attractor with actuator built in."""
    # TODO(jg): How to apply actuator to dynamic system?
    def dy(t, y, *args):
        dy0 = s * (y[1] - y[0])
        dy1 = r * y[0] - y[1] - y[0] * y[2]
        dy2 = y[0] * y[1] - b * y[2] + actuator(*y, *args)
        return [dy0, dy1, dy2]
    return dy


@toolz.curry
def lorenz_in_2(actuator, s=10.0, r=28.0, b=8.0 / 3.0):
    """Return lorenz attractor with actuator built in."""
    # TODO(jg): How to apply actuator to dynamic system?
    def dy(t, y, *args):
        dy0 = s * (y[1] - y[0])
        dy1 = r * y[0] - y[1] - y[0] * y[2] + actuator(*y, *args)
        dy2 = y[0] * y[1] - b * y[2]
        return [dy0, dy1, dy2]
    return dy


# Coupled Systems

@toolz.curry
def van_der_pol(actuator, sensor=toolz.identity, omega=1.0, a=0.1, b=0.01, A=0.0):
    """Return Van der Pol oscillator with actuator built in."""
    def dy(t, y, *args):
        N = int(len(y) / 2)
        y0, y1 = y[:N], y[N:]
        dy0 = y1
        dy1 = - omega**2 * y0 + a * y1 * (1 - b * y0**2) + A.dot(y1) + actuator(*sensor(y), *args)
        return numpy.hstack((dy0, dy1))
    return dy


# TODO(jg): untested
@toolz.curry
def fitzhugh_nagumo(actuator, sensor=toolz.identity, a=0.7, b=0.8, tau=12.5, A=0.0):
    """Return FitzHugh-Nagumo oscillator with actuator built in."""
    if tau == 0.0:
        raise RuntimeError('Division by zero for tau = {}'.format(tau))

    def dy(t, y, *args):
        N = int(len(y) / 2)
        y0, y1 = y[:N], y[N:]
        dy0 = y0 - y0**3 / 3.0 - y1
        dy1 = (y0 + a - b * y1) / tau + A.dot(y1) + actuator(*sensor(y), *args)
        return numpy.hstack((dy0, dy1))
    return dy


# TODO(jg): untested
@toolz.curry
def hindmarsh_rose(actuator, sensor=toolz.identity, a=1.0, b=3.0, c=1.0, d=5.0, r=1e-3, s=4.0, xR=-8.0/5.0, A=0.0):
    """Return Hindmarsh-Rose oscillator with actuator built in."""
    def dy(t, y, *args):
        N = int(len(y) / 3)
        y0, y1, y2 = y[:N], y[N:2*N], y[2*N:]
        dy0 = y1 - a * y0**3 + b * y0**2 - y2
        dy1 = c - d * y0**2 - y1 + A.dot(y1) + actuator(*sensor(y), *args)
        dy2 = r * (s * (y0 - xR) - y2)
        return numpy.hstack((dy0, dy1, dy2))
    return dy


def global_coupling(N):
    """Generate a coupling matrix for global coupling.

    N = 3: A = [-2  1  1]
               [ 1 -2  1]
               [ 1  1 -2]
    """
    A = numpy.ones((N, N))
    numpy.fill_diagonal(A, -1.0 * float(N - 1))
    return A


def pairwise_coupling(N):
    """Generate a coupling matrix for pairwise coupling.

    N = 2: A = [-1  1]
               [ 1 -1]
    """
    assert N % 2 == 0
    a = numpy.array([[-1, 1],
                     [1, -1]])
    A = numpy.kron(numpy.eye(int(N / 2)), a)
    return A


def circular_array_coupling(N):
    """Generate a coupling matrix for circular array coupling.

    N = 4: A = [-2  1  0  1]
               [ 1 -2  1  0]
               [ 0  1 -2  1]
               [ 1  0  1 -2]
    """
    g = networkx.cycle_graph(N)
    A = -1.0 * networkx.linalg.laplacian_matrix(g)
    return A


def grid_2d_coupling(n, m, periodic=False):
    """Generate a coupling matrix for a 2D grid of n*m oscillators.

    n denotes the number of rows and m the number of columns in the grid.
    """
    g = networkx.grid_2d_graph(n, m, periodic=periodic)
    A = -1.0 * networkx.linalg.laplacian_matrix(g, nodelist=sorted(g.nodes()))
    return A


def dorogovtsev_goltsev_mendes_coupling(n):
    """Generate a coupling matrix from the dorogovtsev-goltsev-mendes graph.

    n is the generation.
    """
    g = networkx.generators.dorogovtsev_goltsev_mendes_graph(n)
    A = -1.0 * networkx.linalg.laplacian_matrix(g)
    return A



"""Provide Individual class for gp."""

import sys
import abc
import deap.gp
import deap.base
import sympy
import functools
import itertools
import numpy as np


def sympy_primitive_set(categories=('algebraic', 'trigonometric', 'exponential'), arguments=['y_0'], constants=[], arity=0):
    """Create a primitive set with sympy primitves.

    :param categories: an optional list of function categories for the primitive set. The following are available
                       'algebraic', 'trigonometric', 'exponential', 'exponential', 'logarithm', 'sqrt'.
    """
    pset = deap.gp.PrimitiveSet('main', arity=arity)
    if 'algebraic' in categories:
        pset.addPrimitive(sympy.Add, arity=2)
        pset.addPrimitive(sympy.Mul, arity=2)
        pset.addPrimitive(functools.partial(sympy.Mul, -1.0), arity=1, name='Neg')
    if 'trigonometric' in categories:
        pset.addPrimitive(sympy.sin, arity=1)
        pset.addPrimitive(sympy.cos, arity=1)
    if 'exponential' in categories:
        pset.addPrimitive(sympy.exp, arity=1)
    if 'logarithm' in categories:
        pset.addPrimitive(sympy.ln, arity=1)
    if 'sqrt' in categories:
        pset.addPrimitive(sympy.sqrt, arity=1)

    # Use sympy symbols for argument representation.
    for symbol in itertools.chain(arguments, constants):
        pset.addTerminal(sympy.Symbol(symbol), name=symbol)
    # Dirty hack to make constant optimization possible.
    pset.args = arguments
    pset.consts = constants
    return pset


def sympy_phenotype(individual):
    """Compile python function from individual.

    Uses sympy's lambdify function. Terminals from the primitive set will be
    used as parameters to the constructed lambda function; primitives (like
    sympy.exp) will be converted into numpy expressions (eg. numpy.exp).
    """
    # args = ','.join(terminal.name for terminal in individual.terminals)
    pset = individual.pset
    args = ','.join(arg for arg in itertools.chain(pset.args, pset.consts))
    expr = sympy.sympify(individual.compile())
    func = sympy.utilities.lambdify(args, expr, modules='numpy')
    return func


class Symc(deap.gp.Ephemeral):
    func = staticmethod(lambda: 1.0)
    ret = deap.gp.__type__


def numpy_primitive_set(arity, categories=('algebraic', 'trigonometric', 'exponential', 'symc')):

    pset = deap.gp.PrimitiveSet("main", arity)
    # Use primitive set built-in for argument representation.
    pset.renameArguments(**{'ARG{}'.format(i): 'x{}'.format(i) for i in range(arity)})
    if 'symc' in categories:
        pset._add(Symc)
        pset.terms_count += 1

    def close_function(func, value):
        @functools.wraps(func)
        def closure(*args):
            res = func(*args)
            if isinstance(res, np.ndarray):
                res[np.isinf(res)] = value
                res[np.isnan(res)] = value
            elif np.isinf(res) or np.isnan(res):
                res = value
            return res
        return closure

    if 'algebraic' in categories:
        pset.addPrimitive(np.add, 2, name="Add")
        pset.addPrimitive(np.subtract, 2, name="Sub")
        pset.addPrimitive(np.multiply, 2, name="Mul")
        _div = close_function(np.divide, 1)
        pset.addPrimitive(_div, 2, name="Div")
    if 'trigonometric' in categories:
        pset.addPrimitive(np.sin, arity=1)
        pset.addPrimitive(np.cos, arity=1)
    if 'exponential' in categories:
        pset.addPrimitive(np.exp, arity=1)
        _log = close_function(np.log, 1)
        pset.addPrimitive(_log, arity=1)
    if 'sqrt' in categories:
        _sqrt = close_function(np.sqrt, 1)
        pset.addPrimitive(_sqrt, arity=1)
    if "constants" in categories:
        pset.addTerminal(np.pi, name="pi")
        pset.addTerminal(np.e, name="e")

    nan_context = {'nan': np.nan, 'inf': np.inf}
    pset.context.update(nan_context)

    return pset


class AExpressionTree(deap.gp.PrimitiveTree):
    """Abstract base class for the genotype.

    Derived classes need to specify a primitive set from which the expression
    tree can be build, as well as a phenotype method.
    """

    pset = None
    hasher = str

    def __init__(self, content):
        if self.pset is None:
            raise TypeError("Cannot instantiate abstract {} with abstract attribute pset.".format(self.__class__))
        super(AExpressionTree, self).__init__(content)
        self.fitness = Measure()

    def __repr__(self):
        """Symbolic representation of the expression tree."""
        repr = ''
        stack = []
        for node in self:
            stack.append((node, []))
            while len(stack[-1][1]) == stack[-1][0].arity:
                prim, args = stack.pop()
                repr = prim.format(*args)
                if len(stack) == 0:
                    break
                stack[-1][1].append(repr)
        return repr

    def __hash__(self):
        return hash(self.hasher(self))

    def __eq__(self, other):
        return hash(self) == hash(other)

    def compile(self):
        """Compile the individual's representation into python code.

        Code taken from deap.gp.compile.

        :returns: either a lambda function if the primitive set has one or more
                  arguments defined, or just the evluated function body
                  otherwise.
        """
        code = repr(self)
        if len(self.pset.arguments) > 0:
            args = ",".join(arg for arg in self.pset.arguments)
            code = "lambda {}: {}".format(args, code)
        try:
            return eval(code, self.pset.context, {})
        except MemoryError:
            _, _, traceback = sys.exc_info()
            raise MemoryError('Error in tree evaluation : '
                              'Python cannot evaluate a tree higher than 90.').with_traceback(traceback)

    @property
    def terminals(self):
        """Return terminals that occur in the expression tree."""
        return [primitive for primitive in self if primitive.arity == 0]

    @classmethod
    def from_string(cls, string):
        return super(AExpressionTree, cls).from_string(string, cls.pset)

    @classmethod
    def create_population(cls, size, gen_method=deap.gp.genHalfAndHalf, min=1, max=4):
        """Create a list of individuals of class Individual."""
        if size < 0:
            raise RuntimeError('Cannot create population of size {}'.format(size))
        toolbox = deap.base.Toolbox()
        toolbox.register("expr", gen_method, pset=cls.pset, min_=min, max_=max)
        toolbox.register("individual", deap.tools.initIterate, cls, toolbox.expr)
        toolbox.register("population", deap.tools.initRepeat, list, toolbox.individual)
        return toolbox.population(n=size)


class ANDimTree(list):

    def __init__(self, trees):
        super().__init__(trees)
        self.dim = len(trees)
        self.fitness = Measure()

    @property
    def base(self):
        pass

    def compile(self):
        funcs = [tree.compile() for tree in self]
        return lambda *x: [func(*x) for func in funcs]

    def __repr__(self):
        return str([str(tree) for tree in self])

    @classmethod
    def create_individual(cls, ndim):
        return cls(cls.base.create_population(ndim))

    @classmethod
    def create_population(cls, size, ndim):
        return [cls.create_individual(ndim) for _ in range(size)]

    @property
    def height(self):
        return len(self)


class Measure(deap.base.Fitness):
    """
    This is basically a wrapper around deap.base.Fitness.

    It provides the following enhancements over the base class:
    - more adequate naming
    - copy constructable
    - no weight attribute
    """

    weights = itertools.repeat(-1)

    def __init__(self, values=()):
        if len(values) > 0:
            self.values = values

    def __iter__(self):
        return iter(self.values)

    def __len__(self):
        return len(self.values)

    def get_values(self):
        return tuple(-1.0 * v for v in self.wvalues)

    def set_values(self, values):
        self.wvalues = tuple(-1.0 * v for v in values)

    def del_values(self):
        self.wvalues = ()

    values = property(get_values, set_values, del_values)



from . import argparse
from . import logging
from . import numeric

from functools import wraps

def memoize(obj):
    cache = obj.cache = {}

    @wraps(obj)
    def memoizer(*args, **kwargs):
        key = str(args) + str(kwargs)
        if key not in cache:
            cache[key] = obj(*args, **kwargs)
        return cache[key]
    return memoizer


class Memoize:
    """Memoize(fn) - an instance which acts like fn but memoizes its arguments
       Will only work on functions with non-mutable arguments
       http://code.activestate.com/recipes/52201/
    """
    def __init__(self, fn):
        self.fn = fn
        self.memo = {}
    def __call__(self, *args, **kwargs):
        key = str(args) + str(kwargs)
        if key not in self.memo:
            self.memo[key] = self.fn(*args, **kwargs)
        return self.memo[key]



"""Collection of helper functions for arparse."""

import argparse


def positive_int(string):
    """Check whether string is an integer greater than 0."""
    try:
        value = int(string, base=10)
    except ValueError:
        raise argparse.ArgumentTypeError("invalid int value: '{}'".format(string))
    if value < 1:
        raise argparse.ArgumentTypeError("int value hast to be greater then 0: '{}'".format(string))
    return value


def non_negative_int(string):
    """Check whether string is an integer greater than -1."""
    try:
        value = int(string, base=10)
    except ValueError:
        raise argparse.ArgumentTypeError("invalid int value: '{}'".format(string))
    if value < 0:
        raise argparse.ArgumentTypeError("int value hast to be either 0 or greater then 0: '{}'".format(string))
    return value


def unit_interval(string):
    """Check whether string is a float in the interval [0.0, 1.0]."""
    try:
        value = float(string, base=10)
    except ValueError:
        raise argparse.ArgumentTypeError("invalid float value: '{}'".format(string))
    if value < 0.0 and value > 1.0:
        raise argparse.ArgumentTypeError("float value hast to be in unit interval [0,1]: '{}'".format(string))
    return value


def ntuple(n, to_type=float):
    def evaluate(string):
        """Check whether string is an n-tuple."""
        try:
            value = tuple(to_type(val) for val in string.split(','))
            if len(value) != n:
                raise ValueError
        except ValueError:
            raise argparse.ArgumentTypeError("Must be a {}-tuple of type {}".format(n, to_type))
        return value
    return evaluate



import itertools
import functools
import toolz
import numpy
import scipy.integrate
import scipy.optimize
import scipy.signal
import deap.gp


@toolz.curry
def row(n, array):
    """Take the nth row from array."""
    return array[n, :]


def integrate(dy, yinit, x, f_args=(), integrator='dopri5', **integrator_args):
    """
    Convenience function for odeint().

    Uselful if you do not want to step through the integration, but rather get
    the full result in one call.

    :param dy: callable(x, y, *args)
    :param yinit: sequence of initial values.
    :param x: sequence of x values.
    :param *f_args: (optional) extra arguments to pass to function.
    :returns: y(x)
    """
    res = odeint(dy, yinit, x, f_args=f_args, integrator=integrator, **integrator_args)
    y = numpy.vstack(res).T
    if y.shape[1] != x.shape[0]:
        y = numpy.empty((y.shape[0], x.shape[0]))
        y[:] = numpy.NAN
    if y.shape[0] == 1:
        return y[0, :]
    return y


def odeint(dy, yinit, x, f_args=(), integrator='dopri5', **integrator_args):
    """
    Integrate the initial value problem (dy, yinit).

    A wrapper around scipy.integrate.ode.

    :param dy: callable(x, y, *args)
    :param yinit: sequence of initial values.
    :param x: sequence of x values.
    :param *f_args: (optional) extra arguments to pass to function.
    :yields: y(x_i)
    """
    @functools.wraps(dy)
    def f(x, y):
        return dy(x, y, *f_args)
    ode = scipy.integrate.ode(f)
    ode.set_integrator(integrator, **integrator_args)
    # ode.set_f_params(*f_args)
    # Seems to be buggy! (see https://github.com/scipy/scipy/issues/1976#issuecomment-17026489)
    ode.set_initial_value(yinit, x[0])
    yield ode.y
    for value in itertools.islice(x, 1, None):
        ode.integrate(value)
        if not ode.successful():
            raise StopIteration
        yield ode.y


def rms(y):
    """Root mean square."""
    return numpy.sqrt(numpy.mean(numpy.square(y)))


def strict_subtract(x, y):
    try:
        if x.shape != y.shape:
            raise ValueError('operands could not be broadcast together with shapes {} {}'.format(x.shape, y.shape))
    except AttributeError:
        pass
    return x - y


def rmse(x, y):
    """Root mean square error."""
    return rms(strict_subtract(x, y))


def nrmse(x, y):
    """Normalized, with respect to x, root mean square error."""
    diff = strict_subtract(x, y)
    return rms(diff) / (numpy.max(x) - numpy.min(x))


def cvrmse(x, y):
    """Coefficient of variation, with respect to x, of the rmse."""
    diff = strict_subtract(x, y)
    return rms(diff) / numpy.mean(x)


def silent_numpy(func):
    @functools.wraps(func)
    def closure(*args, **kwargs):
        with numpy.errstate(all='ignore'):
            return func(*args, **kwargs)
    return closure


# this is some legacy code to optimize constants

def generate_context(pset, data):
    context = {arg: dat for arg, dat in zip(pset.arguments, data.T)}
    context.update(pset.context)

    return context


def optimize_constants(ind, cost, context, precision=3, options=None, constraints=None):
    """ Update the constant values of ind according to:
    vec(c) = argmin_c ||yhat(data,c) - y||

    This needs to be called together when using symbolic constants.
    It may be called as a mutation operator together with the usage of ercs.
    """
    idx = [index for index, node in enumerate(ind) if isinstance(node, deap.gp.Ephemeral)]

    if len(idx) == 0:
        return ind

    values = [ind[i].value for i in idx]
    args = [("c%i" % i) for i in range(len(idx))]

    code = str(ind)
    for i, arg in zip(idx, args):
        code = code.replace(ind[i].format(), arg, 1)
    code = "lambda {args}: {code}".format(args=",".join(args), code=code)
    yhat = eval(code, context, {})
    with numpy.errstate(invalid='ignore', over='ignore'):
        res = scipy.optimize.minimize(cost, values, args=yhat, options=options, constraints=constraints)

    if res.success and all(numpy.isfinite(res.x)):
        values = res.x

    for i, value in zip(idx, values):
        ind[i] = type(ind[i])()   # re-initialize the class, else it will overwrite the parents values too!
        ind[i].value = round(value, precision)

        del ind.fitness.values

    return ind



import os
import json
import logging
import random
import argparse
import copy
from functools import partial

import zmq
import yaml
import deap.tools
import deap.gp
import toolz
import numpy as np

from glyph.gp import AExpressionTree
from glyph.utils import Memoize
import glyph.application


class RemoteApp(glyph.application.Application):
    def run(self, break_condition=None):
        """For details see application.Application.
        Will checkpoint and close zmq connection on keyboard interruption.
        """
        try:
            super().run(break_condition=break_condition)
        except KeyboardInterrupt:
            self.checkpoint()
        finally:
            self.assessment_runner.send(dict(action="SHUTDOWN"))
            zmq.Context.instance().destroy()

    @classmethod
    def from_checkpoint(cls, file_name, send, recv):
        """Create application from checkpoint file."""
        cp = glyph.application.load(file_name)
        gp_runner = cp['runner']
        gp_runner.assessment_runner = RemoteAssessmentRunner(send, recv, max_steps=cp['args'].hill_steps, directions=cp['args'].directions,
                                         consider_complexity=cp['args'].consider_complexity, precision=cp['args'].precision, caching=cp['args'].caching)
        app = cls(cp['args'], cp['runner'], file_name)
        app.pareto_fronts = cp['pareto_fronts']
        app._initialized = True
        random.setstate(cp['random_state'])
        return app

    def checkpoint(self):
        """Checkpoint current state of evolution."""

        runner = copy.deepcopy(self.gp_runner)
        del runner.assessment_runner
        glyph.application.safe(self.checkpoint_file, args=self.args, runner=runner,
                               random_state=random.getstate(), pareto_fronts=self.pareto_fronts)
        self.logger.debug('Saved checkpoint to {}'.format(self.checkpoint_file))


def get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument('--port', type=int, default=5555, help='Port for the zeromq communication (default: 5555)')
    parser.add_argument('--ip', type=str, default="localhost", help='IP of the client (default: localhost)')

    config = parser.add_argument_group('config')
    group = config.add_mutually_exclusive_group()
    group.add_argument('--remote', action='store_true', dest='remote', default=False, help='Request GP configs from experiment handler.')
    group.add_argument('--cfile', dest='cfile', type=argparse.FileType('r'), help='Read GP configs from file')

    RemoteApp.add_options(parser)
    cp_group = parser.add_mutually_exclusive_group(required=False)
    cp_group.add_argument('--ndim', type=int, default=1)
    cp_group.add_argument('--resume', dest='resume_file', metavar='FILE', type=str,
                          help='continue previous run from a checkpoint file')
    cp_group.add_argument('-o', dest='checkpoint_file', metavar='FILE', type=str,
                          default=os.path.join('.', 'checkpoint.pickle'),
                          help='checkpoint to FILE (default: ./checkpoint.pickle)')
    parser.add_argument('--verbose', '-v', dest='verbosity', action='count', default=0,
                        help='set verbose output; raise verbosity level with -vv, -vvv, -vvvv')
    parser.add_argument('--logging', '-l', dest='logging_config', type=str, default='logging.yaml',
                        help='set config file for logging; overides --verbose (default: logging.yaml)')
    glyph.application.AlgorithmFactory.add_options(parser.add_argument_group('algorithm'))
    group_breeding = parser.add_argument_group('breeding')
    glyph.application.MateFactory.add_options(group_breeding)
    glyph.application.MutateFactory.add_options(group_breeding)
    glyph.application.SelectFactory.add_options(group_breeding)
    glyph.application.CreateFactory.add_options(group_breeding)

    ass_group = parser.add_argument_group('assessment')
    ass_group.add_argument('--directions', type=int, default=5, help='Number of directions to try in stochastic hill climber (default: 5)')
    ass_group.add_argument('--hill_steps', type=int, default=5, help='Number of iterations of stochastic hill climber (default: 5)')
    ass_group.add_argument('--consider_complexity', type=bool, default=True, help='Consider the complexity of solutions for MOO (default: True)')
    ass_group.add_argument('--caching', type=bool, default=True, help='Cache evaluation (default: True)')
    ass_group.add_argument('--precision', type=int, default=3, help='Precision of constants (default: 3)')
    return parser


def _send(socket, msg, serializer=json):
    socket.send(serializer.dumps(msg).encode('ascii'))


def _recv(socket, serializer=json):
    return serializer.loads(socket.recv().decode('ascii'))


def connect(ip, port):
    socket = zmq.Context.instance().socket(zmq.REQ)
    socket.connect('tcp://{ip}:{port}'.format(ip=ip, port=port))
    send = partial(_send, socket)
    recv = partial(_recv, socket)
    return send, recv


def update_namespace(ns, up):
    """Update the argparse.Namespace ns with a dictionairy up.
    """
    return argparse.Namespace(**{**vars(ns), **up})


def handle_gpconfig(config, send, recv):
    """Will try to load config from file or from remote and update the cli/default config accordingly.
    """
    if config.cfile:
        gpconfig = yaml.load(config.cfile)
    elif config.remote:
        send(dict(action="CONFIG"))
        gpconfig = recv()
    else:
        gpconfig = {}
    return update_namespace(config, gpconfig)


def build_pset_gp(primitives):
    """Build a primitive set used in remote evaluation. Locally, all primitives correspond to the id() function.
    """
    pset = deap.gp.PrimitiveSet('main', arity=0)
    pset.constants = set()
    for fname, arity in primitives.items():
        if arity > 0:
            func = lambda *args: args
            pset.addPrimitive(func, arity, name=fname)
        elif arity == 0:
            pset.addTerminal(fname, name=fname)
            pset.arguments.append(fname)
        else:
            pset.addTerminal(fname, name=fname)
            pset.constants.add(fname)
    if len(pset.terminals) == 0:
        raise RuntimeError("Pset needs at least one terminal node. You may have forgotten to specify it.")
    return pset


class hashabledict(dict):
    """We can use this as dict key"""
    def __hash__(self):
        return hash(tuple(sorted(self.items())))


def default_constants(individual, default=1):
    """Finds all constants which are used in the individual and tries to inherit old values (from parents).
    If a constant cannot be inherited, the default value will be used as initial guess.
    """
    constants_in_ind = {k for k in individual.base.pset.constants if any(k in str(i) for i in individual)}
    old_values = getattr(individual, "constants", {})
    return hashabledict({k: old_values.get(k, default) for k in constants_in_ind})  # try hotstart = inherited values


class RemoteAssessmentRunner:
    def __init__(self, send, recv, consider_complexity=True, max_steps=5, directions=5, caching=True, precision=3):
        """Contains assessment logic. Uses zmq connection to request evaluation.
        Constant optimization is done using a stochastic hill climber.
        """
        self.send = send
        self.recv = recv
        self.consider_complexity = consider_complexity
        self.max_steps = max_steps
        self.directions = directions
        self.precision = precision
        if caching:
            self.evaluate = Memoize(self.evaluate)

    def evaluate(self, individual, constants=None):
        """Evaluate a single individual.
        """
        constants = constants or {}
        self.evaluations += 1
        payload = [str(t) for t in individual]
        for k, v in constants.items():
            payload = [s.replace(k, str(v)) for s in payload]
        self.send(dict(action="EXPERIMENT", payload=payload))
        error = self.recv()["fitness"]
        return error

    def hill_climb(self, individual, rng=np.random):
        """Stochastic hill climber for constant optimization.
        Try self.directions different solutions per iteration to select a new best individual.
        This iterates self.max_steps times.
        """
        constants = default_constants(individual)
        memory = {constants: self.evaluate(individual, constants)}
        if len(constants.keys()) == 0:
            return self.evaluate(individual, constants)

        for _ in range(self.max_steps):
            for _ in range(self.directions):
                c = toolz.valmap(lambda x: tweak(x, self.precision), constants, factory=hashabledict)
                error = self.evaluate(individual, c)
                memory[c] = error
            constants = min(memory, key=memory.get)  # argmin for dictionaries
            memory = {constants: memory[constants]}
        individual.constants = constants

        return memory[constants]

    def measure(self, individual):
        """Construct fitness for given individual.
        """
        error = self.hill_climb(individual)
        if self.consider_complexity:
            fitness = *error, sum(map(len, individual))
        else:
            fitness = error
        return fitness

    def update_fitness(self, population, map=map):
        self.evaluations = 0
        invalid = [p for p in population if not p.fitness.valid]
        fitnesses = map(self.measure, invalid)
        for ind, fit in zip(invalid, fitnesses):
            ind.fitness.values = fit
        return self.evaluations

    def __call__(self, population):
        return self.update_fitness(population)


def tweak(x, p, rng=np.random):
    """ x = round(x + xi, p) with xi ~ N(0, sqrt(x)+10**(-p))
    """
    return round(x+rng.normal(scale=np.sqrt(abs(x))+10**(-p)), p)


class Individual(AExpressionTree):
    pass


class NDTree(glyph.gp.individual.ANDimTree):
    base = Individual


def make_remote_app():
    parser = get_parser()
    args = parser.parse_args()

    send, recv = connect(args.ip, args.port)
    workdir = os.path.dirname(os.path.abspath(args.checkpoint_file))
    if not os.path.exists(workdir):
        raise RuntimeError('Path does not exist: "{}"'.format(workdir))
    log_level = glyph.utils.logging.log_level(args.verbosity)
    glyph.utils.logging.load_config(config_file=args.logging_config, default_level=log_level, placeholders=dict(workdir=workdir))
    logger = logging.getLogger(__name__)

    if args.resume_file is not None:
        logger.debug('Loading checkpoint {}'.format(args.resume_file))
        app = RemoteApp.from_checkpoint(args.resume_file, send, recv)
    else:
        args = handle_gpconfig(args, send, recv)
        try:
            pset = build_pset_gp(args.primitives)
        except AttributeError:
            raise AttributeError("You need to specify the pset")
        Individual.pset = pset
        mate = glyph.application.MateFactory.create(args, Individual)
        mutate = glyph.application.MutateFactory.create(args, Individual)
        select = glyph.application.SelectFactory.create(args)
        create_method = glyph.application.CreateFactory.create(args, Individual)
        ndmate = partial(glyph.gp.breeding.nd_crossover, cx1d=mate)
        ndmutate = partial(glyph.gp.breeding.nd_mutation, mut1d=mutate)
        ndcreate = lambda size: [NDTree(create_method(args.ndim)) for _ in range(size)]
        NDTree.create_population = ndcreate
        algorithm_factory = partial(glyph.application.AlgorithmFactory.create, args, ndmate, ndmutate, select, ndcreate)
        assessment_runner = RemoteAssessmentRunner(send, recv, max_steps=args.hill_steps, directions=args.directions,
                                                   consider_complexity=args.consider_complexity, precision=args.precision, caching=args.caching)
        gp_runner = glyph.application.GPRunner(NDTree, algorithm_factory, assessment_runner)
        app = RemoteApp(args, gp_runner, args.checkpoint_file)

    return app, args


def main():

    app, args = make_remote_app()
    app.run()

if __name__ == "__main__":
    main()



from .algorithms import *
from .algorithms import all_algorithms
from .individual import AExpressionTree, sympy_primitive_set, sympy_phenotype, numpy_primitive_set
from .breeding import all_mutations, all_crossover



import os
import sys
import logging
import logging.config
import yaml

# module name is in conflict with stdlib and can cause unwanted monkey patching


def load_config(config_file, placeholders={}, default_level=logging.INFO):
    """Load logging configuration from .yaml file."""
    if not sys.warnoptions:
        # Route warnings through python logging
        logging.captureWarnings(True)
    if os.path.exists(config_file):
        with open(config_file, 'rt') as f:
            content = f.read().format(**placeholders)
        config = yaml.load(content)
        logging.config.dictConfig(config)
    else:
        logging.basicConfig(level=default_level)


def log_level(verbosity):
    """Convert numeric verbosity to logging log levels."""
    level = [logging.ERROR, logging.WARNING, logging.INFO, logging.DEBUG]
    verbosity = len(level) - 1 if verbosity >= len(level) else verbosity
    return level[verbosity]



"""Symbolic regression tools."""

import re
import ast
from setuptools import setup, find_packages


_version_re = re.compile(r'__VERSION__\s+=\s+(.*)')

with open('glyph/__init__.py', 'rb') as f:
    version = str(ast.literal_eval(_version_re.search(f.read().decode('utf-8')).group(1)))


def read(fname, split=True):
    with open(fname, 'r') as f:
        content = f.read()
    return content.split('\n') if split else content


setup(
    name='pyglyph',
    version=version,
    author='Markus Abel, Julien Gout, Markus Quade',
    author_email='markus.abel@ambrosys.de, julien.gout@ambrosys.de, markus.quade@ambrosys.de',
    description=__doc__.split('\n'),
    long_description=read('README.md', split=False),
    license=read('LICENCE', split=False),
    keywords='complex systems, control, machine learning, genetic programming',
    url='https://www.github.com/ambrosys/glyph',
    packages=find_packages(exclude=["tests", "doc", "examples"]),
    install_requires=read('requirements-to-freeze.txt'),
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Intended Audience :: Science/Research',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.5',
    ],
    entry_points={
        'console_scripts': [
            'glyph-remote = glyph.cli.glyph_remote:main'
        ]
    },
)
